import unittest

class MappersTest(unittest.TestCase):
    
    def test_pd(self):
        
        # from Quanthon.mappers import pauli_decomposition
        # from Quanthon.algorithms import VQE
        import Quanthon as qt
        import numpy as np
        rng = np.random.default_rng()
        # h = np.array([[1,1],[1,2]])
        h2 = np.array([[1,2,1,2],
                       [2,2,1,1],
                       [1,1,3,2],
                       [2,1,2,4]])

        h2 += h2.T
        eigs = np.zeros(10)
        vqe_eigs = np.zeros(10)
        n = 2
        
        for i in range(10):
            h = np.abs(rng.random((6, 6)))
            h += h.T
            eig_val, _ = np.linalg.eigh(h)
            print(f"eig_val: {min(eig_val)}")
            eigs[i] = min(eig_val)
            qubit_ops = qt.pauli_decomposition(h)
            ansatz = qt.HardwareEfficientAnsatz(3)
            vqe = qt.VQE(ansatz, rng.random(ansatz.n_params) * 2 * np.pi - np.pi)
            min_eig = vqe.minimise_eigenvalue(qubit_ops)[1]
            vqe_eigs[i] = min_eig
            print(min_eig)

        import matplotlib.pyplot as plt
        plt.plot(eigs, label='exact')
        plt.plot(vqe_eigs, label='vqe')
        plt.legend()
        plt.show()
        # h_q_op = qt.pauli_decomposition(h)
        # h2_q_op = qt.pauli_decomposition(h2)

        # print(f"h_q_op: {h_q_op}")
        # print(f"h2_q_op: {h2_q_op}")


    

if __name__ == '__main__':
    unittest.main()


from Quanthon.base import Qubits
from Quanthon.expectation import cal_expectation
from Quanthon.utils import pauli_sum
from Quanthon.algorithms import VQE, AdaptVQE
from Quanthon.mappers import jordan_wigner, pauli_decomposition
from Quanthon.physics import Hamiltonian
from Quanthon.ansatzs import HardwareEfficientAnsatz, QubitAdaptAnsatz


'''Doc:
    Normal VQE, Adapt-VQE
'''

from scipy.optimize import minimize, Bounds
import numpy as np
import warnings
from .expectation import cal_expectation
from .utils import pauli_sum
from .base import Qubits

class VQE():
        def __init__(self, ansatz, init_points, optimiser=minimize):
            '''
            args:
                ansatz: a parametriced circuit that takes parmas: theta and phi
                init_points: a list of initial points, must match the number of the parameters in the ansatz
                expectation: the function that calculates the expectation value of the ansatz'''
            self.ansatz = ansatz
            self.params = init_points # has to match the number of the parameters in the ansatz
            self.expectation = cal_expectation
            try:
                ansatz.create_circuit(init_points)
            except ValueError:
                raise ValueError(f'The initial points ({init_points}) do not match the dimension of the ansatz.')

            self.minimise = optimiser
                

        def _objective(self, params):

            # print(params)
            params = np.real(params)
            self.ansatz.create_circuit(params)
            qc = self.ansatz.qubits
            qc.run()
            
            energy = self.expectation(qc, self.H, self.num_shots)

            # self.H_mat = pauli_sum(self.H)
            # energy = self.ansatz.qubits.state.conj().T @ self.H_mat @ self.ansatz.qubits.state
            return energy

        def minimise_eigenvalue(self, H_pauli_str, num_shots=10000):
            '''
            Rotates the parametrised circuit to find the minimised energy using classical 
            minimisation algorithms.
            Inputs:
            H_pauli_str: the Hamiltonian of the system in transformed in terms of Pauli strings,
            num_shots: (int) number of shots,
            return: (float) minimised energy eigenvalues.'''
            self.H = H_pauli_str

            self.num_shots = num_shots

            lb = -np.pi * np.ones(len(self.params))
            ub = np.pi * np.ones(len(self.params))
            bounds = Bounds(lb, ub)
            
            result = self.minimise(self._objective, 
                                   self.params, 
                                #    method='Powell',
                                #    method='Nelder-Mead',
                                     method='COBYLA',
                                    #  method='TNC',
                                   bounds=bounds, 
                                   options= {"maxiter": 100}
                                   )
            min_params = result.x
            min_energy = result.fun
            
            # print(min_params)
            return min_params, min_energy
    

    
class AdaptVQE():
     
    def __init__(self, ansatz, expectation=None, optimiser=minimize, estimate_energy=True):
        '''
        args:
            ansatz: An instance of QubitAdaptAnsatz to be optimised,
            expectation: the function which estimates the energy given a Hamiltonian,
            optimiser: the minimisation function used, default to be scipy.opmtimize.minimize,
            estimate_energy: if True, uses the estimator and shots, otherwise uses the matrix product 
                             of the state and the Hamiltonian.
        '''
        self.old_params = [] # list of parameters for the adapt circuit 
        # self.old_gates = []
        self.ansatz = ansatz
        self.expectation = cal_expectation

        self.minimise = optimiser
        
    
    def gradient(self, A):
        
        '''
        A: the operator whose gradient is calculated.
        '''
        # in practice one should estimate this as well
        # print(f"params: {self.params}")
        # print(self.ansatz.qubits)
        state = self.ansatz.run_without_update(self.params)
        # print(self.ansatz.qubits)
        
        # apply all gates first

        grad = state.conj().T @ (self.H_mat @ A - A @ self.H_mat) @ state

        
        return grad

    def _append_operator(self, eps=10e-2):
        
        '''
            Append a new operator to the ansatz which has the largest greadient ~ [H, A]
            return:
                True if largest gradient is < eps and a new operator is added, False otherwise.
        '''

        op_cand = self.ansatz.pool[0]
        max_abs_grad = 0

        # print(op_cand)
        
        for op in self.ansatz.pool:
            op_mat = pauli_sum([(op.strip('i'), 1j)])
            op_grad = self.gradient(op_mat)
            abs_grad = np.abs(op_grad)
            
            # print(f"op: {op}, grad: {op_grad}")

            if abs_grad > max_abs_grad:
                op_cand = op
                max_abs_grad = abs_grad
        
        if max_abs_grad < eps:
            print(f"end of adapt, grad = {max_abs_grad}")
            return False # gradient = 0, end of adapt 
        
        # print(f"appendding, op: {op_cand}, grad: {max_abs_grad}")
        # grow the ansatz
        # self.old_gates.append(op_cand) # type(op_cand) == str
        self.params = np.append(self.params, 0)
        self.ansatz.append_op(op_cand)

        return True
    
    def _objective(self, params):
        # print("called _objective")
        state = self.ansatz.run_without_update(params)
        qc = Qubits(self.ansatz.qubits.n_qubit)
        qc.set_state(state)

        energy = self.expectation(qc, self.H_str, self.num_shots)
        # energy = state.conj().T @ self.H_mat @ state
        return energy
    
    def run_adapt_circuit(self, H, num_shots=10000, max_iter=100, grad_eps=1e-4):
        '''
        args:
            H: list of 2-tuple such as [('II', 0.5)]
            num_shots: number of shots,
            max_iter: maximum number of iterations,
        
        return: 
            the adapted ansatz and the energy eigenvalues.
        '''
        self.H_mat = pauli_sum(H)
        self.H_str = H
        self.params = np.array([])
        for i in range(max_iter): 
            # need to update the state too
            # print(self.ansatz.qubits)
            # self.ansatz.run(self.params) # ??????
            old_params, energy = self.minimise_eigenvalue(num_shots) # state is not updated here
            self.params = old_params
            print(f"i: {i}, min_energy = {energy}")
            

            if not self._append_operator(eps=grad_eps): # new parameter is added here 
                break
            
        self.ansatz.run(self.params) # update the state with the final parameters
        return self.ansatz, energy
    

    def minimise_eigenvalue(self, num_shots=10000):

        self.num_shots = num_shots
        
        # print(self.params)
        result = self.minimise(self._objective, self.params, method='Powell', options= {"maxiter": 100000})
        min_params = result.x
        min_energy = result.fun

        # self.ansatz.run(min_params) # update the state using only the optimal parameters
         
        return min_params, min_energy


import numpy as np


def get_ob_op(n, i, j, coeff):
    # when the indices are different

    all_op = []

    if i < j:
        base = ['' if k == i or k == j else 'Z' if i < k < j else 'I' for k in range(n)]
        pauli_terms = [('XX', 0.25), ('XY', 0.25j), ('YX', -0.25j), ('YY', 0.25)]
    else:
        i, j = j, i  # the qubits that's smaller always on the left!!
        base = ['' if k == i or k == j else 'Z' if i < k < j else 'I' for k in range(n)] 
        # can't move outside conditional because i, j are swapped.
        pauli_terms = [('XX', 0.25), ('XY', -0.25j), ('YX', 0.25j), ('YY', 0.25)]

        # all gates acting on smaller numbers always on the left

    for op, factor in pauli_terms:
        new_op = base.copy()
        new_op[i] = op[0]
        new_op[j] = op[1]
        new_cf = factor * coeff
        all_op.append((''.join(new_op), new_cf))
    return all_op


def gen_prod_4u(ops):
    """
    Generate products of creation and annihilation operators in the given order.

    args:
        ops: list of length 4, consists of creation and annihilation operators depending on the order 
        of their indices.
            
    return:
        list of strings, each string is a term of the product
    """

    op1, op2, op3, op4 = ops
    products = []

    for t1, factor_1 in op1:
        for t2, factor_2 in op2:
            for t3, factor_3 in op3:
                for t4, factor_4 in op4:
                    op = t1 + t2 + t3 + t4 
                    factor = factor_1 * factor_2 * factor_3 * factor_4
                    products.append((op, factor))

    return products

def gen_prod_3u(ops):
    """
    Generate products of creation and annihilation operators in the given order.

    args:
        ops: list of length 3, consists of creation and annihilation operators depending on the order 
        of their indices.
            
    return:
        list of strings, each string is a term of the product
    """

    op1, op2, op3 = ops
    products = []

    for t1, factor_1 in op1:
        for t2, factor_2 in op2:
            for t3, factor_3 in op3:
                op = t1 + t2 + t3
                factor = factor_1 * factor_2 * factor_3
                products.append((op, factor))

    return products

def get_tb_base(n,ordered_indices):

    # get the I's and the Z's for the qubits not acted on. 
    tb_op_base = list(n * 'I')

    for i in range(n):
        if i in ordered_indices:
            continue

        if ordered_indices[0] < i < ordered_indices[1]:
            tb_op_base[i] = 'Z'
        
        if ordered_indices[2] < i < ordered_indices[3]:
            tb_op_base[i] = 'Z'
        
    return tb_op_base
    
def get_tb_4u(n,p,q,r,s,coeff):

    '''Get the two body operaors when all 4 indices are unique.'''
    
    ordered_indices = sorted([p,q,r,s])

    tb_op_base = get_tb_base(n, ordered_indices)

    creation = [('X', 0.5), ('Y', -0.5j)]
    annihilation = [('X', 0.5), ('Y', 0.5j)]

    sorted_zip = sorted(zip([p,q,r,s], [creation, creation, annihilation, annihilation]))
    ops = [op for _, op in sorted_zip]

    op_products = gen_prod_4u(ops)
    # print(op_products)

    all_tb_op = []
    for op, factor in op_products:
        # every product turn into one operator in the sum
        if np.allclose(coeff, 0):
            continue # skip coeffs close to 0
        
        new_op = tb_op_base.copy()
        
        for i, index in enumerate(ordered_indices):
            new_op[index] = op[i]
        
        new_op = ''.join(new_op)
        all_tb_op.append((new_op, factor * coeff))
        # print(new_op, coeff)
    return all_tb_op


def get_tb_3u(n, p, q, r, s, coeff):

    '''Get the two body operaors when 3 indices are unique.
    args:
        n: int, number of qubits,
        p, q, r, s: int, the indices of the qubits the second quantisation operation is on.
    
    return:
        all_tb_op: list of tuples in the form of '''
    # print(p, q, r, s)
    ordered_indices = sorted([p,q,r,s])

    tb_op_base = get_tb_base(n, ordered_indices)

    creation = [('X', 0.5), ('Y', -0.5j)]
    annihilation = [('X', 0.5), ('Y', 0.5j)]
    c_times_a = [('I', 0.5), ('Z', -0.5)]

    appeared = set()
    for i in [p,q,r,s]:    
        if i in appeared:
            repeated_index = i
        else:
            appeared.add(i)

    ops = []
    ordered_indices.remove(repeated_index) 
    for i in ordered_indices: # now len 3
        if i == repeated_index:
            ops.append(c_times_a)
        elif i == p or i == q:
            ops.append(creation)
        else:
            ops.append(annihilation)

    op_products = gen_prod_3u(ops)

    all_tb_op = []
    for op, factor in op_products:
        # every product turn into one operator in the sum
        if np.allclose(coeff, 0):
            continue # skip coeffs close to 0
        
        new_op = tb_op_base.copy()
        
        for i, index in enumerate(ordered_indices):
            new_op[index] = op[i] # op is a pauli string for the indices given not the whole qubit oprator
        
        new_op = ''.join(new_op)
        all_tb_op.append((new_op, factor * coeff))
        # print(new_op, coeff)

    return all_tb_op 


def check_health_jw(h_pauli, n):
    count = 0
    for term in h_pauli:
        if len(term[0]) != n:
            print(term)
            count += 1
    
    return count
    

def simplify_pauli_terms(terms):

    '''
        Simplify the pauli sums by collecting like terms.
        
        
        return: list, pauli sum equal to the input with like terms collected and sorted.
    
    '''
    simplified_terms = {}
    
    for pauli_string, coeff in terms:
        if pauli_string in simplified_terms:
            simplified_terms[pauli_string] += coeff
        else:
            simplified_terms[pauli_string] = coeff
    
    simplified_terms_list = [(pauli_string, coeff) for pauli_string, coeff in simplified_terms.items() if coeff != 0]
    
    # will sort alphabetically
    return sorted(simplified_terms_list)


def is_power_of_two(n):
    return n > 0 and (2 ** (n.bit_length() - 1)) == n


def expand_h_mat(h_mat):

    '''For matrices that are not a power of 2, expand the matrix to the nearest power of 2.
    
    args:
        h_mat: 2d numpy array, the matrix to be expanded.

    return:
        new_h_mat: 2d numpy array, the expanded matrix to size of smallest power of 2 possible. 
        The original matrix is at the top left corner.
    '''

    n = h_mat.shape[0]
    pow = 1
    while True:
        if n <= 2 ** pow:
           break 
        pow += 1
    
    new_h_mat = np.zeros((2 ** pow, 2 ** pow))
    new_h_mat[:n, :n] = h_mat
    for i in range(n, 2 ** pow):
        new_h_mat[i, i] = 99
    # print(new_h_mat)
    return new_h_mat
 




'''Doc:
    Ansatz for VQE calculations. Can grow if needs be. Parametrise the qubits object.
    parameters are taken as arguments to the circuit at every iteration'''

# TODO: add init_state to HardwareEfficientAnsatz too

# DEBUGGER

def qprint(msg):
    
    print("QubitAdaptAnsatz")
    print(msg)


import numpy as np
from scipy.linalg import expm


from .base import Qubits, Gate
from .utils import pauli_sum

class Ansatz:

    '''Should work for any type of ansatz that are not evolving, do not use on its own.'''
    def __init__(self, n_qubits, reps=1) -> None:

        self.n_qubits = n_qubits
        self.reps = reps
        self.n_params = 2 * n_qubits * reps

  
class HardwareEfficientAnsatz(Ansatz):

    def __init__(self, n_qubits, reps=1) -> None:
        super().__init__(n_qubits, reps)


    def create_circuit(self, init_params):
        
        self.qubits = Qubits(self.n_qubits)
        if len(init_params) != self.n_params:
            raise ValueError(f'Initial parameters do not match the number of parameters required for the ansatz: {len(init_params)} != {self.n_params}')
        
        reshaped_params = init_params.reshape(self.reps, 2*self.n_qubits)
        
        for r in range(self.reps):
            for i in range(self.n_qubits):
                # check if the parameters are real
                if reshaped_params[r, i] != reshaped_params[r, i].real:
                    raise ValueError(f'Parameter {[r,i]} is not real.')
                self.qubits.Rx(reshaped_params[r, i], i)
                self.qubits.Ry(reshaped_params[r, i + self.n_qubits], i)
        
        for r in range(self.reps):
            for i in range(self.n_qubits):
                if i != self.n_qubits - 1:
                    self.qubits.CNOT(i, i+1)

        # self.qubits.draw()

    def run(self):
        self.qubits.run()
    


class QubitAdaptAnsatz:

    def __init__(self, n_qubits, pool='V', init_state=None) -> None:

        '''Ansatz for the Adapt-VQE calculation.'''

        self.qubits = Qubits(n_qubits)

        if init_state is None:
            # initial state not sepcified, initialise randomly
            init_state = np.random.rand(2**n_qubits) + 1j * np.random.rand(2**n_qubits)
            init_state = init_state / np.linalg.norm(init_state) 
        
        self.init_state = init_state
        self.qubits.set_state(init_state)

        if pool == 'V':
            self.pool = self.create_complete_V_pool(n_qubits)
        
        elif pool == 'G':
            self.pool = self.create_complete_G_pool()
        
        else:
            self.pool = pool # custom pool, doens't have to be complete
            
        

    def __repr__(self) -> str:
        
        return f"QubitAdaptAnsatz of {self.qubits} qubits and pool: {self.pool}."
    
    def create_complete_V_pool(self, n):
        
        if n == 2:
            return ['iYZ', 'iIY']

        prev_set = self.create_complete_V_pool(n - 1)
        new_pool = [] # set()

        for prev_op in prev_set:
            new_op = prev_op + 'Z'
            new_pool.append(new_op)


        new_pool.append('i' + (n-1) * 'I' + 'Y')
        new_pool.append('i' + (n-2) * 'I' + 'YI')


        return new_pool
        

    def create_complete_G_pool(self):
        raise NotImplementedError('Not implemented yet')


    def append_op(self, op):

        '''
        op: string, representing one of the operators in the pool
        
        '''
        # no longer need to append all the gates since they are now saved

        qprint(f"append_op op: {op}")
        op_mat = pauli_sum([(op.strip('i'), 1j)])

        def parametrised_mat(param, op_mat):
            return expm(param * op_mat)
        
        # self.params.append(0) # the corresponding parameter for the gate, initialised to 0.
        
        adapt_gate = Gate(f'exp({op})', matrix=lambda param: parametrised_mat(param, op_mat), n_qubits=self.qubits.n_qubit)
        self.qubits.circuit.append(adapt_gate)
        
    
    def run(self, params):
        for i, gate in enumerate(self.qubits.circuit):
            # print(gate.matrix(params[i]))
            self.qubits.state = gate.act(self.qubits.state, params[i])
        
        # print(self.qubits.circuit)
    
    def run_without_update(self, params):
        '''Run the circuit without updating the state.'''
        old_state = self.qubits.state
        for i, gate in enumerate(self.qubits.circuit):
            self.qubits.state = gate.act(self.qubits.state, params[i])
        
        new_state = self.qubits.state
        self.qubits.state = old_state # return to the new state while returning the result of the circuit
        return new_state



    

        



    





import unittest
import numpy as np

# testing features
from Quanthon import VQE, AdaptVQE, jordan_wigner, Hamiltonian, QubitAdaptAnsatz, HardwareEfficientAnsatz, pauli_sum

# test tools
from .test_cmp_qk import qiskit_hardware
from .test_utils import is_hermitian


class AlgorithmTest(unittest.TestCase):


    def test_vqe(self):

        op_str =[("II", -1.052373245772859),
                    ("IZ", 0.39793742484318045),
                    ("ZI", -0.39793742484318045),
                    ("ZZ", -0.01128010425623538),
                    ("XX", 0.18093119978423156),
                ]
        
        ansatz = HardwareEfficientAnsatz(2, reps=2)

        rng = np.random.default_rng(826)
        n_params = ansatz.n_params

        init_points = rng.random(n_params) * 2 * np.pi - np.pi 
        vqe = VQE(ansatz, init_points)
        min_params, min_energy = vqe.minimise_eigenvalue(op_str, 1000)
        print(min_energy)

        # with qiskit

        result = qiskit_hardware(op_str)
        # print("diff from qiskit")
        print(result.eigenvalue)
    
    def _test_adapt_vqe(self):

        qubit_op =[("II", -1.052373245772859),
                    ("IZ", 0.39793742484318045),
                    ("ZI", -0.39793742484318045),
                    ("ZZ", -0.01128010425623538),
                    ("XX", 0.18093119978423156),
                ]
        h_mat = pauli_sum(qubit_op)
        # print(f"h_mat: {h_mat}")

        ansatz = QubitAdaptAnsatz(2)
        print(ansatz.qubits)

        print(ansatz)
        

        vqe = AdaptVQE(ansatz)
        ansatz, min_energy = vqe.run_adapt_circuit(qubit_op, 10000, max_iter=10)
        print(min_energy)

        # with qiskit

        result = qiskit_hardware(qubit_op)
        print(result.eigenvalue)

    def _test_pairing(self):
            
        qubit_op = [('IIIIII', 0.1875), ('IIIIIZ', -0.5625), 
                    ('IIIIZI', -0.0625), ('IIIZII', 0.4375), 
                    ('IIZIII', -0.5625), ('IIZIIZ', 0.0625), 
                    ('IXXIXX', 0.03125), ('IXXIYY', (0.03125+0j)), 
                    ('IXYIXY', (-0.03125+0j)), ('IXYIYX', (0.03125+0j)), 
                    ('IYXIXY', (0.03125+0j)), ('IYXIYX', (-0.03125+0j)), 
                    ('IYYIXX', (0.03125+0j)), ('IYYIYY', (0.03125+0j)), 
                    ('IZIIII', -0.0625), ('IZIIZI', 0.0625), 
                    ('XXIXXI', 0.03125), ('XXIYYI', (0.03125+0j)), 
                    ('XYIXYI', (-0.03125+0j)), ('XYIYXI', (0.03125+0j)), 
                    ('XZXXZX', 0.03125), ('XZXYZY', (0.03125+0j)), 
                    ('XZYXZY', (-0.03125+0j)), ('XZYYZX', (0.03125+0j)), 
                    ('YXIXYI', (0.03125+0j)), ('YXIYXI', (-0.03125+0j)), 
                    ('YYIXXI', (0.03125+0j)), ('YYIYYI', (0.03125+0j)), 
                    ('YZXXZY', (0.03125+0j)), ('YZXYZX', (-0.03125+0j)), 
                    ('YZYXZX', (0.03125+0j)), ('YZYYZY', (0.03125+0j)), 
                    ('ZIIIII', 0.4375), ('ZIIZII', 0.0625)]
        

        h_mat = pauli_sum(qubit_op)
        if not is_hermitian(h_mat):
            raise Warning('This Hamiltonian is not hermitian.')

        my_result = self.my_hardware(qubit_op)
        print(my_result)

        result = qiskit_hardware(qubit_op)
        print(result.eigenvalue)

    def my_hardware(self, qubit_op, seed=267):
        n = len(qubit_op[0][0])
        ansatz = HardwareEfficientAnsatz(n, reps=1)
        rng = np.random.default_rng(seed)
        n_params = ansatz.n_params
        init_points = rng.random(n_params) * 2 * np.pi - np.pi
        # print(init_points)
        vqe = VQE(ansatz, init_points)
        min_params, min_energy = vqe.minimise_eigenvalue(qubit_op, 10000)
        print(min_energy.real)


if __name__ == '__main__':
    unittest.main()



class Hamiltonian:

    def __init__(self, ob_coeffs, tb_coeffs) -> None:
        '''
        args:
            ob_coeffs: 2d array of one body coefficients.
            
            tb_coeffs: 4d array of two body coefficients.
        '''
        self.one_body_coeffs = ob_coeffs 
        self.two_body_coeffs = tb_coeffs

    def __repr__(self) -> str:
        return f'Hamiltonian: 1b {self.one_body_coeffs.shape}, 2b {self.two_body_coeffs.shape}.'

 


from itertools import product
import numpy as np


def get_all_pauli(n):

	'''Returns a list of all possible Pauli strings of length n (number of qubits)'''
	pauli = 'IXYZ'
	# all_paulis = list(product(pauli, pauli, repeat=int(n/2)))
	perms = [''.join(p) for p in product(pauli, repeat=n)]
	return perms


def _no_pauli_after_i(pauli_str):

	detected_I = False
	for op in pauli_str:
		if not detected_I:
			if op == 'I':
				detected_I = True
				continue
			else:
				continue
		if op != 'I':
			return False
	return True

def get_SWAPs(pauli_str, SWAP_list):

	if _no_pauli_after_i(pauli_str):
		return ''.join(pauli_str), SWAP_list

	found_I = False
	pauli_str = list(pauli_str)
	non_i_indx = None
	for i, op in enumerate(pauli_str):
		if op == 'I':
			found_I = True
			I_indx = i
			continue
		if found_I:
			if op != 'I':
				non_i_indx = i
			pauli_str[I_indx] = op 
			pauli_str[non_i_indx]= 'I'
			SWAP_list.append((I_indx, non_i_indx))
			found_I = False

	return get_SWAPs(pauli_str, SWAP_list)

def get_CNOTs(pauli_str): 

	if not _no_pauli_after_i:
		raise ValueError('Pauli string must not have any Pauli operator after I.')
	CNOT_pairs = []

	for i in range(len(pauli_str)-1):
		if pauli_str[i+1] == 'I':
			break
		CNOT_pairs.append((i+1, i))
		
	return CNOT_pairs

def change_basis(pauli_str):

	for i in pauli_str:
		if i not in 'IXYZ':
			raise ValueError(f'Invalid Pauli operator: {i}')

	SWAPed_pauli, SWAPs = get_SWAPs(pauli_str, [])
	CNOT_pairs = get_CNOTs(SWAPed_pauli)

	change_basis_op = []

	for pair in SWAPs:
		change_basis_op.append(('SWAP', pair))
	
	# print(SWAPed_pauli)
	for i, op in enumerate(SWAPed_pauli):
		if op == 'X':
			change_basis_op.append(('H', i))
		elif op == 'Y':
			change_basis_op.append(('Sdag', i))
			change_basis_op.append(('H', i))
		else:
			change_basis_op.append(('I', i))
 
	
	for pair in reversed(CNOT_pairs):
		change_basis_op.append(('CNOT', pair))
	
	return change_basis_op

def rotate_basis(qc, ops):
	'''rotate to the measurement basis'''
	
	# print("bc", sum(qc.state**2))

	for op_type, idx in ops:
		if op_type == 'CNOT':
			qc.CNOT(idx[0], idx[1])
		elif op_type == 'SWAP':
			qc.SWAP(idx[0], idx[1])
		elif op_type == 'X':
			qc.X(idx)
		elif op_type == 'Y':
			qc.Y(idx)
		elif op_type == 'Z':
			qc.Z(idx)
		elif op_type == 'H':
			qc.H(idx)
		elif op_type == 'Sdag':
			qc.Sdag(idx)
		elif op_type == 'I':
			continue
		else:
			raise ValueError(f'Invalid operator: {op_type}')
	
	qc.run()
	# print([(gate.name, gate.n_qubits) for gate in qc.circuit])
	# print('ops', ops)
	# print([gate.name for gate in qc.circuit])
	# for gate in qc.circuit:
	#     qc.state = gate.act(qc.state)
		# print(gate.name)
		# print(qc.state)
		# print(qc.state**2)
		# print(sum(qc.state**2))
	
	
	
def cal_expectation(qc, pauli_ops, n_shots=10000):
	
	'''
	return:
		the expectation value
	args:
		qc: the circuit to be measured in
	'''

	expectation = 0
	for pauli_str, coeff in pauli_ops:

		
		if set(pauli_str) == {'I'}:
			expectation += coeff
			continue
		# print(pauli_str, coeff)
		state_eigval = np.ones(len(qc.state))
		state_eigval[int(0.5*len(qc.state)):] *= -1 # first half of state has eigenvalue 1 and the second half -1
		# print(state_eigval)
		qc_copy = qc.copy()
		cb_ops = change_basis(pauli_str)
		# print(cb_ops)

		rotate_basis(qc_copy, cb_ops)

		# print(qc_copy)
		counts = qc_copy.measure(n_shots)[:, 0]
		# print("this count", counts)
		# print(counts)
		expectation += coeff * np.sum(counts * state_eigval) / n_shots
	
	# print(cb_ops)
	return expectation


if __name__ == '__main__':
	# Test Pauli operators
	
	# print(get_all_pauli(2))
	# pauli_str = 'YY'
	# for op in pauli_str:
	#     rotation_ops = get_rotations(op)
	#     print(rotation_ops)

	a = 'X'
	b = 'Y'
	c = 'IZ'
	d = 'ZIXX'
	# test no_op_after_i
	
	# print(_no_pauli_after_i(a), _no_pauli_after_i(b), _no_pauli_after_i(c))

	# print(_no_pauli_after_i(d))

	# PASSED

	# test get_SWAPs
	
	# for i in [a, b, c, d]:
	#     result, SWAPs = get_SWAPs(i, [])
	#     result = "".join(result)
	#     print(result, SWAPs)
	
	# PASSED

	# test get_CNOTs

	# for paulis in [a, b, c, d]:
	#     result, SWAPs = get_SWAPs(paulis, [])
	#     result = "".join(result) 
	#     print(result)
	#     CNOT_pairs = get_CNOTs(result)
	#     print(f"CNOTs: {paulis}", CNOT_pairs)

	# PASSED

	# test change_basis

	# for i in [a, b, c, d]:
	#     big_op = change_basis(i)
	#     print(f"big_op for {i}", big_op)

	# PASSED

	# test expectation

	# a = 'X'
	# b = 'IYYI'
	# c = 'ZIII'
	# d = 'XIXI'

	# b = 'XX'
	# c = 'YY'
	# d = 'IZ'
	# e = 'XY'

	# testing drawing, seems ok

	# qc = Qubits(2)
	# qc.H(0)
	# qc.CNOT(0,1)
	# qc.run()
	# # print(qc)

	# hamiltonian_str = [b, c, e]
	# coeffs = {b:0.5 , c:0.5 , d:0.5, e:3}
	# ham = [(op, coeffs[op]) for op in hamiltonian_str]
	
	# # print(ham)

	# print(f"hamiltonian: {hamiltonian_str}")


	# # print(ham_mat)
	# # for term in hamiltonian:
		
	# #     qc_copy = qc.copy()
	# #     print(qc_copy)
	# #     ops = change_basis(term)
	# #     rotate_basis(qc_copy, ops)
	# #     print(ops)
	# #     print(qc_copy.gate_history)
	# #     qc_copy.draw()

	# # testing expectation
	
	# for term in hamiltonian_str:
		
	#     qc_copy = qc.copy()
	#     # print(qc_copy)
	#     ops = change_basis(term)
	#     rotate_basis(qc_copy, ops)
	#     # print(ops)
	
	# estimated_energy = expectation(qc_copy, ham, n_shots=100000)
	# print(estimated_energy)

	# from Quanthon import pauli_sum
	# # ham_mat = pauli_sum(ham)
	# # real_energy = qc.state.conj() @ ham_mat @ qc.state
	# # print(real_energy.real)

	# # Testing expectation 2

	# qc = Qubits(4)
	# qc.H(0)
	# qc.CNOT(0,1)
	# qc.CNOT(1,2)
	# qc.CNOT(2,3)
	# qc.run()
	# Ham = [
	#     # ('ZIZI', (1)),
	#     ('ZZXX', (0.5)), 
	#     # ('YYII', (1)),
	#     # ('IXIX', (0.5))
	#     ]

	# H_mat = pauli_sum(Ham)
	# ee = expectation(qc, Ham, n_shots=100000)
	# re = qc.state.conj().T @ H_mat @ qc.state

	# print("exact: ", re, "expectation: ", ee)
 





from qiskit_nature.second_q.mappers import JordanWignerMapper
from qiskit_nature.second_q.operators import FermionicOp

from mappers import jordan_wigner
from physics import Hamiltonian

import numpy as np

def matrix_to_fermionic_op(one_body_matrix, two_body_matrix):
    # Get the size of the one-body matrix
    n = one_body_matrix.shape[0]

    # Initialize an empty FermionicOp dictionary
    op_dict = {}

    # Handle one-body terms
    for i in range(n):
        for j in range(n):
            if one_body_matrix[i, j] != 0:
                key = f"+_{i} -_{j}"
                op_dict[key] = one_body_matrix[i, j]

    # Handle two-body terms
    # Assuming two_body_matrix is a 4-dimensional array
    for i in range(n):
        for j in range(n):
            for k in range(n):
                for l in range(n):
                    if two_body_matrix[i, j, k, l] != 0:
                        # Create the corresponding fermionic operator for the two-body term
                        key = f"+_{i} +_{j} -_{k} -_{l}"
                        op_dict[key] = two_body_matrix[i, j, k, l]

    # Create and return the FermionicOp
    return FermionicOp(op_dict)


# Convert the matrix to FermionicOp




a=np.complex64(1.5)
b=np.complex64(2+1j)
c=np.complex64(2)

# one_body = np.zeros((4,4), dtype=np.complex_)
# np.fill_diagonal(one_body, [c,a])
# # one_body[0,2] = b
# # one_body[2,0] = b.conj()
# h = Hamiltonian(one_body, np.zeros((4,4,4,4)))

# # rng = np.random.default_rng()
# # a = np.complex64(rng.random() + 1j*rng.random())
# # b = np.complex64(rng.random() + 1j*rng.random())
# # c = np.complex64(rng.random() + 1j*rng.random())

# # h = Hamiltonian(np.array([[a, b], [b.conj(), c]]), np.zeros((2,2,2,2)))
# k = dict(jordan_wigner(h))

# fermionic_op = matrix_to_fermionic_op(one_body)

# # op = FermionicOp({
# #     "+_0 -_0": c,
# #     "+_0 -_1": b.conj(),
# #     "+_1 -_0": b,
# #     "+_1 -_1": a
# # })

# op = matrix_to_fermionic_op(one_body)

# mapper = JordanWignerMapper()
# q = mapper.map(second_q_ops=op)


def test_1():

    h = Hamiltonian(np.array([[a, b], [b.conj(), c]]), np.zeros((2,2,2,2)))
    k = dict(jordan_wigner(h))
    op = FermionicOp({
        "+_0 -_0": c,
        "+_0 -_1": b.conj(),
        "+_1 -_0": b,
        "+_1 -_1": a
    })

    mapper = JordanWignerMapper()
    q = mapper.map(second_q_ops=op)


    d = {"II":0.5*a+0.5*c, 
        "IZ":-0.5*c, 
        "XX":0.25*(b+b.conj()), 
        "XY":0.25j*(b-b.conj()), 
        "YX": 0.25j*(b.conj()-b), 
        "YY":0.25*(b+b.conj()), 
        "ZI":-0.5*a}

    print("Pauli, Qiskit, Paper, Quanthon")
    for i,p in enumerate(q.paulis.to_labels()):
        print(p, q.coeffs[i], d[p], k[p])
        np.testing.assert_equal(q.coeffs[i], d[p])
        np.testing.assert_equal(q.coeffs[i], k[p])


    print(len(k) == len(d))

def test_one_body():

    n = 4
    a=np.complex64(1.5)
    b=np.complex64(2+1j)
    c=np.complex64(2)

    one_body = np.zeros((n,n), dtype=np.complex_)
    np.fill_diagonal(one_body, [a,c])
    one_body[0,2] = b
    one_body[2,0] = b.conj()
    one_body[1,3] = b
    one_body[3,1] = b.conj()
   
    # h = Hamiltonian(np.array([[a, b], [b.conj(), c]]), np.zeros((2,2,2,2)))
    
    op = matrix_to_fermionic_op(one_body)

    h = Hamiltonian(np.flip(one_body), np.zeros((n,n,n,n))) 
    # flip cuz qiskit is weird

    k = dict(sorted(jordan_wigner(h)))
    mapper = JordanWignerMapper()
    q = mapper.map(second_q_ops=op)

    print("Pauli, Qiskit, Quanthon")
    for i,p in enumerate(q.paulis.to_labels()):
        print(p, q.coeffs[i], k[p]) 
        np.testing.assert_equal(q.coeffs[i], k[p])


    print(f"my dict: {len(k)}, qiskit: {len(q.coeffs)}")
    # for i in k:
    #     print(i, k[i])

def test_two_body():
    n = 4

    a,b,c = np.array([2,8,6], dtype=np.complex_)
    b = np.complex_(8+1j)
    c = np.complex_(6)
    # one_body = np.zeros((n,n))
    one_body = np.random.rand(n,n) + 1j*np.random.rand(n,n)
    one_body = one_body + one_body.conj().T

    two_body = np.random.rand(n,n,n,n) + 1j*np.random.rand(n,n,n,n)
    two_body = two_body + two_body.conj().T

    # # diagonal
    # two_body[0,1,0,1] = a
    # two_body[0,2,2,0] = b
    # two_body[2,0,0,2] = b
    # two_body[1,3,3,1] = c
    # two_body[3,1,1,3] = c

    # # 4 unique
    # two_body[0,1,2,3] = b # ijkl
    # two_body[2,3,0,1] = b.conj() # klij
    # two_body[2,3,1,0] = c
    # two_body[1,0,2,3] = c.conj()

    # # 3 unique
    # two_body[0,1,1,3] = b
    # two_body[1,3,0,1] = b.conj()


    op = matrix_to_fermionic_op(one_body,two_body)

    h = Hamiltonian(np.flip(one_body), np.flip(two_body))

    return h, op

debug = True

if debug:
    # test_1()
    # test_one_body()
    h, op = test_two_body()

    k = dict(sorted(jordan_wigner(h)))

    mapper = JordanWignerMapper()

    q = mapper.map(second_q_ops=op)
    # print(q)
    # print(k)

    print("Pauli, Qiskit, Quanthon")
    for i,p in enumerate(q.paulis.to_labels()):
        print(p, q.coeffs[i], k[p]) 
        assert(np.allclose(q.coeffs[i], k[p]))
        # np.testing.assert_equal(q.coeffs[i], k[p])

    # for key in k.keys():
    #     if key not in q.paulis.to_labels():
    #         print(f"Key {key} not in qiskit, has value {k[key]}")

    print(f"my dict: {len(k)}, qiskit: {len(q.coeffs)}")


from qiskit_nature.second_q.mappers import JordanWignerMapper
from qiskit_nature.second_q.operators import FermionicOp
from qiskit_nature.second_q.algorithms import GroundStateEigensolver

from qiskit.quantum_info import SparsePauliOp
from qiskit.primitives import Estimator
from qiskit.circuit.library import TwoLocal

from qiskit_algorithms import VQE
from qiskit_algorithms import AdaptVQE

from qiskit_algorithms.optimizers import COBYLA, L_BFGS_B, SLSQP
from qiskit_algorithms.utils import algorithm_globals

import warnings

import numpy as np

def qiskit_hardware(op):
        
        warnings.filterwarnings("ignore")

        estimator = Estimator()
        qubit_op = SparsePauliOp.from_list(op)

        # we will iterate over these different optimizers
        optimizers = [COBYLA(maxiter=80), L_BFGS_B(maxiter=60), SLSQP(maxiter=60)]
        converge_counts = np.empty([len(optimizers)], dtype=object)
        converge_vals = np.empty([len(optimizers)], dtype=object)

        for i, optimizer in enumerate(optimizers):
            print("\rOptimizer: {}        ".format(type(optimizer).__name__), end="")
            algorithm_globals.random_seed = 50
            ansatz = TwoLocal(rotation_blocks="ry", entanglement_blocks="cz")

            counts = []
            values = []

            def store_intermediate_result(eval_count, parameters, mean, std):
                counts.append(eval_count)
                values.append(mean)

            vqe = VQE(estimator, ansatz, optimizer, callback=store_intermediate_result)
            result = vqe.compute_minimum_eigenvalue(operator=qubit_op)
            converge_counts[i] = np.asarray(counts)
            converge_vals[i] = np.asarray(values)
        
        return result

def _qiskit_adapt(self):
    
    pass
    warnings.filterwarnings("ignore")

    estimator = Estimator()
    qubit_op = SparsePauliOp.from_list(
    [
        ("II", -1.052373245772859),
        ("IZ", 0.39793742484318045),
        ("ZI", -0.39793742484318045),
        ("ZZ", -0.01128010425623538),
        ("XX", 0.18093119978423156),
    ]
    )

    # we will iterate over these different optimizers
    optimizer = SLSQP(maxiter=60)

    print("\rOptimizer: {}        ".format(type(optimizer).__name__), end="")
    algorithm_globals.random_seed = 50
    ansatz = TwoLocal(rotation_blocks="ry", entanglement_blocks="cz")

    vqe = VQE(estimator, ansatz, optimizer)
    adapt_vqe = AdaptVQE(vqe)
    eigenvalue, _ = adapt_vqe.compute_minimum_eigenvalue(qubit_op)
    print(eigenvalue)

    return eigenvalue

def matrix_to_fermionic_op(one_body_matrix, two_body_matrix):
    # Get the size of the one-body matrix
    n = one_body_matrix.shape[0]

    # Initialize an empty FermionicOp dictionary
    op_dict = {}

    # Handle one-body terms
    for i in range(n):
        for j in range(n):
            if one_body_matrix[i, j] != 0:
                key = f"+_{i} -_{j}"
                op_dict[key] = one_body_matrix[i, j]

    # Handle two-body terms
    # Assuming two_body_matrix is a 4-dimensional array
    for i in range(n):
        for j in range(n):
            for k in range(n):
                for l in range(n):
                    if two_body_matrix[i, j, k, l] != 0:
                        # Create the corresponding fermionic operator for the two-body term
                        key = f"+_{i} +_{j} -_{k} -_{l}"
                        op_dict[key] = two_body_matrix[i, j, k, l]

    # Create and return the FermionicOp
    return FermionicOp(op_dict)

def make_random_hamiltonian(n=4, seed=42):

    # one_body = np.zeros((n,n))
    rng = np.random.default_rng(seed)
    one_body = rng.random((n,n)) + 1j*rng.random((n,n))
    one_body = one_body + one_body.conj().T # make it hermitian

    two_body = rng.random((n,n,n,n)) + 1j*rng.random((n,n,n,n))
    two_body = two_body + two_body.conj().T

    op = matrix_to_fermionic_op(one_body,two_body)

    h = Hamiltonian(np.flip(one_body), np.flip(two_body))

    # h: Hamiltonian, op: FermionicOp for qiskit stuff
    return h, op


import numpy as np
from itertools import product
from .mappers_utils import *
from .utils import get_all_paulis, is_hermitian, get_pauli



def jordan_wigner(hamiltonian):
    
    '''
    Perform the Jordan-Wigner transform for which maps second quantisation Hamiltonian to qubit Hamiltonians.
    N.B. 0th qubit to the left, always.
    arg:
        hamiltonian: an instance of the Hamiltonian class, the second quantisation Hamiltonian containing the overlap integrals.
    
    return:
        h_pauli: list of tuples, the qubit Hamiltonian in terms of Pauli strings.'''
    h_pauli = []
    n = hamiltonian.one_body_coeffs.shape[0]

    # one body
    for i in range(n):
        for j in range(n):

            if np.allclose(hamiltonian.one_body_coeffs[i, j],0):
                continue

            # diagonal terms
            if i == j:
                # II...III - I...IZ...I

                all_i = n * 'I'
                h_pauli.append((all_i, 0.5 * hamiltonian.one_body_coeffs[i, j]))

                z_term = list(all_i)
                # z_term[n-i-1] = 'Z'
                z_term[i] = 'Z'
                z_term = ''.join(z_term)

                h_pauli.append((z_term, -0.5 * hamiltonian.one_body_coeffs[i, j]))

            else:
                ob_op = get_ob_op(n, i, j, hamiltonian.one_body_coeffs[i, j])
                h_pauli.extend(ob_op)
     
                
    # two body
                
    # print("unhealthy terms after one body", _check_health_jw(h_pauli, n))
    # return h_pauli

    for p in range(n):
        for q in range(n):
            if p == q:
                continue
            for r in range(n):
                for s in range(n):
                    if r == s:
                        continue


                    coeff = hamiltonian.two_body_coeffs[p, q, r, s]

                    if np.allclose(coeff,0):
                        continue
                    
                    # print(f'pqrs: {p}{q}{r}{s}, coeff: {coeff}')

                    
                    # if the outer indices are not all greater than the inner indices
                    if (p > q and r > s) or (p < q and r < s):
                        coeff *= -1
                        # print(coeff)

                    if (p == r and q == s) or (p == s and q == r):

                        # print(p, q, r, s)
                        all_i = n*'I'

                        pz = list(n*'I')
                        pz[p] = 'Z'
                        pz = ''.join(pz)

                        qz = list(n*'I')
                        qz[q] = 'Z'
                        qz = ''.join(qz)

                        qpz = list(n*'I')
                        qpz[p] = 'Z'
                        qpz[q] = 'Z'
                        qpz = ''.join(qpz)

                        # print(p, q, r, s, coeff)
                        # print(all_i, pz, qz, qpz)
                        # print()
                        # print("all_i", all_i)
                        # print("pz", pz)
                        # print("qz", qz)
                        # print("qpz", qpz)

                        h_pauli.append((all_i, 0.25 * coeff))
                        h_pauli.append((pz, -0.25 * coeff))
                        h_pauli.append((qz, -0.25 * coeff))
                        h_pauli.append((qpz, 0.25 * coeff))

                        
                    elif len({p, q, r, s}) == 4:
                        # if they are all different

                        tb_op = get_tb_4u(n, p, q, r, s, coeff)
                        h_pauli.extend(tb_op)

                    elif len({p, q, r, s}) == 3:
                        # 0223
                        tb_op = get_tb_3u(n, p, q, r, s, coeff)
                        h_pauli.extend(tb_op)


    if check_health_jw(h_pauli, n):
        raise ValueError("There are terms whose length is not equal to the number of qubits.")
    
    h_pauli = simplify_pauli_terms(h_pauli)
    return h_pauli



def pauli_decomposition(h_mat):

    '''
    Perform the Pauli decomposition of a Hermitian matrix of size (2^n,2^n).
    args:
        h_mat: 2d array, the Hamiltonian matrix to be decomposed.
    
    return:
        h_pauli: list of tuples, the qubit Hamiltonian in terms of Pauli strings.
    
    ''' 

    # checks
    if not is_hermitian(h_mat):
        raise ValueError("The Hamiltonian matrix is not Hermitian.")
    
    h_pauli = []
    
    if not is_power_of_two(h_mat.shape[0]):
        h_mat = expand_h_mat(h_mat)
        
    n = int(np.log2(h_mat.shape[0]))
    all_paulis = get_all_paulis(n)
    
    for paulis in all_paulis:
        coeff = (1/2**n) * np.trace(get_pauli(paulis) @ h_mat)
        if coeff != 0:
            h_pauli.append((paulis, coeff))
    
    
    return h_pauli


# Under construction
'''This will be the last thing to be implemented with built in models, Hamiltonians and special mapping if any.'''
import numpy as np
from itertools import permutations


class Model():

    def __init__(self, integral_values, mapper='jw') -> None:
        '''
        args:
            integral_values: array of (*, 2) of integral values for the model.
        '''
        self.s = 'ad,a'
        
        mapper_dict = {'jw': self.jordan_wigner,
                       'other': self.other_mapper,
                       'special': self.special_mapper}
        
        self.states = np.zeros(np.log2(self.hamiltonian.shape[0]))

        if mapper not in self.mapper_dict.keys():
            self.mapper = mapper
        else:
            raise ValueError(f"Unknown mapper {mapper}.")
    
    # def hamiltonian(self, pq_max):
    #     assert(type(pq_max) == int)
    #     range_of_pq = np.arange(0, pq_max)
        

    def adag(self, i):
        ''' applies the creation operator on state i.
        N.B. This state is the energy state of the model, not the quantum state used in quantum computers.'''
        
        if self.states[i] == 1:
            raise ValueError(f"State {i} is already occupied.")
        self.states[i] = 1


    def a(self,i):
        ''' applies the annihilation operator on state i'''

        if self.states[i] == 0:
            raise ValueError(f"Nothing to annihilate at state {i}.")
        self.states[i] = 0

    def jordan_wigner(self):

        return 
    
    def other_mapper(self):
        raise NotImplementedError('Other mappers are not implemented yet.')

        
    def _to_Pauli_strings(self, mapper='jw'):
        if self._special_mapping_exist:
            self._special_mapping()  
            
        pass

    def estimator(self):
        ''' Genrate the correct rotation to do the measurements for Hamilronians given by the pauli strings.
        
        return:
            (list) of the rotations 
        '''

        pass

class Lipkin_Model():  
    '''The Lipkin (LMG) model.''' 
    #TODO: combine into Model at some point.
    
    def __init__(self, J=1, mapper='jw') -> None:
        super().__init__(mapper)
        self.J = J
        self.hamiltonian = self.Lipkin_Hamiltonian(J)
        self._special_mapping_exist = True

    def Lipkin_Hamiltonian(v, J=1, z_coeff=0.5):
        
        '''
        returns a string of pauli operator given the J values of the lipkin model.'''
        x_coeff = -v/2
        y_coeff = v/2

        pauli_str = []
        Z_str = 'I'*(2*J-1) + 'Z'
        Z_ops = set(permutations(Z_str))
        # print(Z_ops)
        Z_ops = [''.join(p) for p in Z_ops]
        for op in Z_ops:
            # print(op)
            pauli_str.append((op, z_coeff))

        # the X's   
        X_str = 'I'*(2*J-2) + 'XX'
        X_ops = set(permutations(X_str))
        X_ops = [''.join(p) for p in X_ops]

        for op in X_ops:
            # print(op)
            pauli_str.append((op, x_coeff))

        # the Y's 

        Y_str = 'I'*(2*J-2) + 'YY'
        Y_ops = set(permutations(Y_str))
        Y_ops = [''.join(p) for p in Y_ops]

        for op in Y_ops:
            # print(op)
            pauli_str.append((op, y_coeff))
        
        return pauli_str


class Ising(Model):

    def __init__(self, mapper='jw') -> None:
        super().__init__(mapper)


class Heisenberg(Model):

    def __init__(self, mapper='jw') -> None:
        super().__init__(mapper)
    
    
class SSH(Model):

    def __init__(self, mapper='jw') -> None:
        super().__init__(mapper)


class KitaevChain(Model):

    def __init__(self, mapper='jw') -> None:
        super().__init__(mapper)






from qiskit_nature.second_q.drivers import PySCFDriver
from qiskit_nature.second_q.mappers import JordanWignerMapper

import numpy as np

def get_h2():

    driver = PySCFDriver(atom="H 0 0 0; H 0 0 0.7414", charge=0, spin=0, basis='sto-3g')
    problem = driver.run()

    hamiltonian = problem.hamiltonian.second_q_op()

    h = np.zeros((4,4))
    u = np.zeros(((4,4,4,4)))
    for label, coeff in sorted(hamiltonian.items()):
        label = remove_non_num(label)
        if len(label) == 2:
            h[int(label[0]), int(label[1])] = coeff
        elif len(label) == 4:
            u[int(label[0]), int(label[1]), int(label[2]), int(label[3])] = coeff
        else:
            raise ValueError("Label not recognized")
        # print(f"{coeff:+.8f} * '{label}'")

    return h, u

def remove_non_num(string):

    for char in string:
        if char not in '0123456789':
            string = string.replace(char, '')
    return string


if __name__ == "__main__":
    h, u  = get_h2()

    u_list = []
    for i in range(4):
        for j in range(4):
            for k in range(4):
                for l in range(4):
                    # if u[i,j,k,l] != 0:
                    #     print(f'{i}{j}{k}{l}, {u[i,j,k,l]}')
                    # if len({i,j,k,l}) == 1:
                    #     continue
                    
                    if (i == k and j == l) or (i == l and j == k):
                        # print(i, j, k, l)
                        u_list.append(u[i,j,k,l])

    u_list = 0.25 * np.array(u_list)
    print(sum(u_list))

    # qiskit get this -0.81261796

    h_coeff = h.diagonal()
    h_coeff = 0.5 * h_coeff
    
    print(sum(h_coeff))

    print(sum(u_list) + sum(h_coeff))


'''The basic classes for quantum computing. Replaced base.py, which was discontinued after Quanthon 0.3.0'''

# TODO: Add n dimension rotation 

import numpy as np
from collections import Counter
from .utils import one_fixed_bit, flip_bit, swap_bits

# Constants
rng = np.random.default_rng()
rangle = '\u27E9'

class Gate:

    def __init__(self, name, matrix, n_qubits):
        '''
        args:
            name: string, the name of the gate;
            matrix: operator matrix or a function which takes params is an argument and returns the matrix of the correct size;
            n_qubits: int, the number of qubits the gate acts on;
            '''
        self.name = name
        self.matrix = matrix

        self.n_qubits = n_qubits
        # self.params = params


    def __repr__(self):
        return f"Gate: {self.name} \n Matrix: \n {self.matrix} \n"
    
    def _check_is_unitary(self, matrix):

        if not np.allclose(matrix @ matrix.conj().T, np.eye(matrix.shape[0])):
            raise ValueError(f"{self.name} is not unitary.")
    
    def act(self, state, param=None):
        if param is not None:
            self._check_is_unitary(self.matrix(param))
            return self.matrix(param) @ state
        self._check_is_unitary(self.matrix)
        return self.matrix @ state

    
class Qubits:

    def __init__(self,n):
        self.state = np.zeros(2**n, dtype=np.complex_)
        self.state[0] = 1
        self.n_qubit = n
        self.opeartor_size = 2**n

        self.I = np.eye(2)
        self.h = 1/np.sqrt(2) * np.array([[1, 1], [1, -1]])
        self.x = np.array([[0, 1], [1, 0]])
        self.y = np.array([[0, -1j], [1j, 0]])
        self.z = np.array([[1, 0], [0, -1]])
        self.s = np.array([[1, 0], [0, 1j]])

        self.circuit = []

        self._get_state_dict()
        self._get_gate_history()
        
    def __repr__(self) -> str:
        return f"Qubit(s) in state: \n {self._to_comp_basis()} \n"

    def _get_gate_history(self):
        self.gate_history = {}
        for i in range(self.n_qubit):
            self.gate_history[f'{i}'] = []
        # print(self.n_qubit)
        # print(self.gate_history)
    
    def _update_gate_history(self, gate, i):
        
        if gate in ['H', 'X', 'Y', 'Z', 'Sdag']:
            self.gate_history[f'{i}'].append(gate) 
        
        elif gate.startswith('Rx') or gate.startswith('Ry'):
            self.gate_history[f'{i}'].append(gate)
 
        elif gate == 'CNOT':
            self.gate_history[f'{i[0]}'].append(gate + f"ctrl{i[1]-i[0]}") 
            self.gate_history[f'{i[1]}'].append(gate + "trgt")
        
        elif gate == 'SWAP':
            self.gate_history[f'{i[0]}'].append(gate + f"1{i[1]-i[0]}")
            self.gate_history[f'{i[1]}'].append(gate + "2")

        else:
            raise ValueError(f"Invalid gate {gate}")
        
        for j in range(self.n_qubit):
            if isinstance(i, tuple):
                if j not in i:
                    self.gate_history[f'{j}'].append('I')
            else:
                if j != i:
                    self.gate_history[f'{j}'].append('I')


    def set_state(self, state):
        assert len(state) == 2**self.n_qubit, f"Invalid state: must have length {2**self.n_qubit}"
        assert np.isclose(np.linalg.norm(state), 1), "Invalid state: must be normalised."
        self.state = state

    def reset_state(self):
        self.state = np.zeros(2**self.n_qubit, dtype=np.complex_)
        self.state[0] = 1
    
    def reset_circuit(self):
        self.circuit = []

    def copy(self):
        new_qubits = Qubits(self.n_qubit)
        new_qubits.state = self.state.copy()
        return new_qubits


    def _get_state_dict(self):
        self.state_dict = dict()
        for i in np.arange(2 ** self.n_qubit):
            self.state_dict[format(i, f'0{self.n_qubit}b')] = self.state[i]
        
    def _to_comp_basis(self):
        # Calculate the number of qubits'

        # Find the indices of the states with non-zero amplitudes
        non_zero_indices = np.where(np.abs(self.state) > 0)[0]

        # Convert each non-zero index to binary representation, along with the corresponding coefficient
        computational_str = ""
        np.zeros(2**self.n_qubit, dtype=np.complex_)
        for idx in non_zero_indices:
            binary_str = format(idx, f"0{self.n_qubit}b") # set to binary
            amplitude = self.state[idx]
            computational_str += f"{amplitude:.2f}|{binary_str}⟩ + "

        computational_str = computational_str.rstrip("+ ")

        # Return the computational basis notation
        return computational_str
    
    def _make_op_mat(self, op, indx): 
        ''' Operates the qubit with the given operator and index. '''
        result = np.eye(2**indx)
        result = np.kron(result, op)
        # print(op)
        # rest_of_indices = int(self.n_qubit - indx - np.sqrt(len(op)))
        rest_of_indices = int(self.n_qubit - indx - np.log2(len(op)))
        for _ in range(rest_of_indices):
            result = np.kron(result, np.eye(2))
        
        return result
        # self.state = result @ self.state

    def H(self,i):
        # self.operate(self.h,i)
        matrix = self._make_op_mat(self.h, i)
        self.circuit.append(Gate('H', matrix, self.n_qubit))
        self._update_gate_history('H', i)
    
    def X(self,i):
        # self.operate(self.x, i)
        matrix = self._make_op_mat(self.x, i)
        self.circuit.append(Gate('X', matrix, self.n_qubit))
        self._update_gate_history('X', i)

    def Y(self,i):
        # self.operate(self.y, i)
        matrix = self._make_op_mat(self.y, i)
        self.circuit.append(Gate('Y', matrix, self.n_qubit))
        self._update_gate_history('Y', i)

    def Z(self,i):
        # self.operate(self.z, i)
        matrix = self._make_op_mat(self.z, i)
        self.circuit.append(Gate('Z', matrix, self.n_qubit))
        self._update_gate_history('Z', i)
    
    def Sdag(self,i):
        # self.operate(self.s.conj(), i)
        matrix = self._make_op_mat(self.s.conj(), i)
        self.circuit.append(Gate('Sdag', matrix, self.n_qubit))
        self._update_gate_history('Sdag', i)

    def Rx(self, theta, i):

        rx = np.cos(theta/2) * self.I - 1j * np.sin(theta/2) * self.x
        # self.operate(Rx, i) 
        matrix = self._make_op_mat(rx, i)
        self.circuit.append(Gate('Rx', matrix, self.n_qubit))
        self._update_gate_history(f'Rx_{theta}', i)

    def Ry(self, phi, i):
        ry = np.cos(phi/2) * self.I - 1j * np.sin(phi/2) * self.y
        # self.operate(Ry, i)
        matrix = self._make_op_mat(ry, i)
        self.circuit.append(Gate('Ry', matrix, self.n_qubit))
        self._update_gate_history(f'Ry_{phi}', i)

    def CNOT(self, control, target):

        if self.n_qubit == 1:
            raise ValueError("The CNOT gate can not be applied to a single qubit.")
        
        matrix = np.eye(self.opeartor_size)

        indices = one_fixed_bit(self.n_qubit, self.n_qubit - control - 1, is_decimal=True) # we do n-c cuz our 0th bit is on the left

        for i in indices:
            # print(i)
            f = flip_bit(i, self.n_qubit - target - 1) # same reason
            # print(f)
            matrix[i, i] = 0
            matrix[f, i] = 1
            matrix[i, f] = 1

        self.circuit.append(Gate('CNOT', matrix, self.n_qubit))
        # self.state = matrix @ self.state
        self._update_gate_history("CNOT", (control, target))


    def SWAP(self, qubit1, qubit2):
        if self.n_qubit == 1:
            raise ValueError("The SWAP gate can not be applied to a single qubit.")
        

        matrix = np.zeros((self.opeartor_size, self.opeartor_size))
        for i in range(self.opeartor_size):
            j = swap_bits(i, self.n_qubit - qubit1 - 1, self.n_qubit - qubit2 - 1)
            matrix[i, j] = 1
            matrix[j, i] = 1

        # self.state = matrix @ self.state
        self.circuit.append(Gate('SWAP', matrix, self.n_qubit))
        self._update_gate_history("SWAP", (qubit1, qubit2))
    
    def run(self):
        
        '''Execute the circuit. Return the result.'''
        for gate in self.circuit:
            self.state = gate.act(self.state)
        
    def run_and_reset(self):
        '''Execute the circuit and reset the circuit. Return the result.'''
        self.run()
        state = self.state
        self.reset_circuit()
        return state

    def prob(self):
        prob = np.abs(self.state**2)
        return prob

    def measure(self, n_shots=1):
        ''' n: number of shots 
            indexs: the index of the qubit(s) being measured '''
        
        prob = self.prob()
        allowed_outcomes = np.arange(len(self.state))
        # print(self.state)
        outcomes = np.random.choice(allowed_outcomes, p=prob, size = n_shots)
        
        self.state = np.zeros_like(self.state)
        self.state[outcomes[-1]] = 1
        counts = Counter(outcomes)
        
        outcomes_count = np.zeros((len(self.state),2)) # 2: state and count
        for i in range(len(self.state)):
            if i in counts:
                outcomes_count[i] = counts[i], format(i, f"0{self.n_qubit}b")
            else:
                outcomes_count[i] = 0,format(i, f"0{self.n_qubit}b")
        # count_qubit = Counter(outcomes_count,)

        # single_qubit_count = Counter(outcomes_count)

        return outcomes_count
        # return outcomes
    
    def draw(self, use_quantikz=True):
        print("cricuit")
        if not use_quantikz:
            raise NotImplementedError("This feature is not implemented yet.")
        else:
            gate_map = {
                'H': '\\gate{H}',
                'I': '\\qw',
                'X': '\\gate{X}',
                'Y': '\\gate{Y}',
                'Z': '\\gate{Z}',
                'Rx': lambda angle: '\\gate{R_x(' + f"{float(angle).real:.3f}" + ')}',
                'Ry': lambda angle: '\\gate{R_y(' + f"{float(angle).real:.3f}" + ')}',
                'Sdag': '\\gate{S^\\dagger}',
                'CNOTctrl': lambda dist: "\\ctrl{" + f"{dist}" + '}',
                'CNOTtrgt': '\\targ{}',
                'SWAP1': lambda dist: "\\swap{" + f"{dist}" + '}',
                'SWAP2': '\\targX{}',
                # 'exp': lambda axis, theta: f'\\gate{R_{axis}}'
            }

            max_gate_length = max(len(gates) for gates in self.gate_history.values())
            
            quantikz_str = "\\begin{quantikz}\n"
            
            for qubit in range(self.n_qubit):
                gates = self.gate_history.get(str(qubit), [])
                quantikz_str += "\t\\lstick{$q_{" + f"{qubit}" + "}$} & "
                
                for gate in gates:
                    if gate.startswith('CNOTctrl'):
                        # print(gate_map.get('CNOTctrl'))
                        quantikz_str += gate_map.get('CNOTctrl')(gate[8:])+ " & "
                    elif gate.startswith('SWAP1'):
                        quantikz_str += gate_map.get('SWAP1')(gate[5:]) + " & "
                    elif gate.startswith('Rx'):
                        quantikz_str += gate_map.get('Rx')(gate[3:]) + " & "
                    elif gate.startswith('Ry'):
                        quantikz_str += gate_map.get('Ry')(gate[3:]) + " & "
                    else:
                        quantikz_str += gate_map.get(gate, gate) + " & "
                quantikz_str += "\qw \\\\\n"
            
            # Fixing the positions of CNOT gates
            # quantikz_str = quantikz_str.replace('\\ctrl & \\targ', '\\targ & \\ctrl')
            
            quantikz_str += "\\end{quantikz}"
            
            print(quantikz_str)


if __name__ == "__main__":
    qc = Qubits(6)
    # qc.H(1)
    # qc.H(0)

    # qc.CNOT(0,1)
    # qc.CNOT(1,2)
    # qc.CNOT(2,3)


    # qc.run()
    # print(qc)

    # qc.X(1)
    # qc.Y(0)
    # qc.rx(0.4, 1)

    # print(qc.gate_history)
    # qc.draw(True)

    print(qc)


# Utils
import numpy as np
from itertools import product
def get_pauli(basis_name):
    pauli_matrices = {'I': np.eye(2, dtype=np.complex128), 
                      'X': np.array([[0, 1], [1, 0]], dtype=np.complex128), 
                      'Y': np.array([[0, -1j], [1j, 0]]), 
                      'Z': np.array([[1, 0], [0, -1]],dtype=np.complex128)}

    # Check if basis_name is a valid combination
    valid_basis = all(ch in pauli_matrices for ch in basis_name)
    if not valid_basis:
        print(f'Invalid basis name: {basis_name}')
        return

    # Calculate the matrix for the given basis_name
    matrix = np.eye(1)
    for ch in basis_name:
        matrix = np.kron(matrix, pauli_matrices[ch])

    return matrix


def pauli_sum(lst):
    """
    Computes the Pauli sum of a list of Pauli operators. Returns the matrix.
    """
    result = 0+0j
    for item in lst:
        if len(item)!= 2:
            raise ValueError(f'must be of length 2, not {len(item)}')
        
        matrix = get_pauli(item[0])
        if matrix is None:
            raise ValueError(f'Invalid Pauli operator: {item[0]}')
        
        result += np.complex128(item[1]) * matrix 

    return result

def entropy(H, level=0):
    '''computes the von Neumann entropies for subsystem A and subsystem B
    input: 
        H: the Hamiltonian matrix (with lmb already included)
    return:
        level: int, for which energy level to compute the entropy, default is 0
        S_A, S_B, the von Neumann entropies for subsystem A and subsystem B'''
    eigenvalues, eigenvectors = np.linalg.eigh(H)

    def _compute_entropy(rho):
        eigenvalues, _ = np.linalg.eig(rho)
        entropy = -np.sum(eigenvalues * np.log2(eigenvalues + 1e-10))
        return entropy
    # index of the lowest eigenvalue and eigenvector
    
    sorted_index = np.argsort(eigenvalues)
    if type(level) == int:
        level = [level]

    entropy_A = np.zeros(len(level))
    entropy_B = np.zeros(len(level))
    for i in level:
        alpha = eigenvectors[:, sorted_index[i]]

        # density matrix for the lowest energy state
        rho_0 = np.outer(alpha, np.conj(alpha))

        # partial trace of both of the subsystems
        rho_A = np.einsum("ijkl,jl->ik", rho_0.reshape(2, 2, 2, 2), np.eye(2))
        rho_B = np.einsum("ijkl,ik->jl", rho_0.reshape(2, 2, 2, 2), np.eye(2))
        
        entropy_A[i] = _compute_entropy(rho_A)
        entropy_B[i] = _compute_entropy(rho_B)

    # Print the von Neumann entropies for subsystem A and subsystem B
    return entropy_A, entropy_B


def sum_of_pow_2(a, b):
    '''generate a sequence of sums of powers of 2 from a to b'''

    return sum([2**i for i in range(a, b+1)])

def one_fixed_bit(n,c,is_decimal=False):
    '''generate the bit string of length n such that the cth bit is always 1.

    args:
        n: int, length of the bit string.
        c: the index of the bit to be fixed to 1
        
    return:
        combinations: string, list of all the bit stirngs of length n whose cth bit is 1.
    
    N.B. 
        to convert to int, specify int(bit_sting, 2) for binary input.'''

    combinations = []
    for i in range(2 ** n):
        bit_string = f'{i:0{n}b}'
        if bit_string[n-c-1] == '1': # careful '1' is not 1
            if is_decimal:
                combinations.append(i)
            else:
                combinations.append(bit_string)

    return combinations

def flip_bit(i, t):
    '''flip the t'th bit of the binary representation of i.
    args: 
        i: int
        t: int
    
    return:
        int, the new integer after flipping the t'th bit.
    '''

    return i ^ (1 << t)


def swap_bits(val, i, j): # https://stackoverflow.com/questions/12173774/how-to-modify-bits-in-an-integer
    """
    Given an integer val, swap bits in positions i and j if they differ
    by flipping their values, i.e, select the bits to flip with a mask.
    Since v ^ 1 = 0 when v = 1 and 1 when v = 0, perform the flip using an XOR.
    """
    if (val >> i) & 1 != (val >> j) & 1:
        mask = (1 << i) | (1 << j)
        val ^= mask

    return val


def get_all_paulis(n):
    '''
    args: 
        n: the number of qubits.
    
    return: list of all the possible Pauli strings of length n.
    '''
    paulis = 'IXYZ'
    return [''.join(p) for p in product(paulis, repeat=n)] 


def is_hermitian(mat):
    return np.allclose(mat, mat.T.conj())

def is_unitary(mat):
    return np.allclose(mat @ mat.T.conj(), np.eye(len(mat)))

    
if __name__ == "__main__":
    # Test Pauli operators
    # pauli_ops = [('IZZZ', 3), ('ZXYI', 2)]
    qubit_op = [('IIIIII', 0.1875), ('IIIIIZ', -0.5625), 
                    ('IIIIZI', -0.0625), ('IIIZII', 0.4375), 
                    ('IIZIII', -0.5625), ('IIZIIZ', 0.0625), 
                    ('IXXIII', 0.0625), ('IXXIIZ', -0.03125), 
                    ('IXXIZI', -0.03125), ('IXYIIZ', 0.03125j), 
                    ('IXYIZI', -0.03125j), ('IYXIIZ', -0.03125j), 
                    ('IYXIZI', 0.03125j), ('IYYIII', (0.0625+0j)), 
                    ('IYYIIZ', (-0.03125+0j)), ('IYYIZI', (-0.03125+0j)), 
                    ('IZIIII', -0.0625), ('IZIIZI', 0.0625), 
                    ('XXIIII', 0.0625), ('XXIIZI', -0.03125), 
                    ('XXIZII', -0.03125), ('XYIIZI', 0.03125j), 
                    ('XYIZII', -0.03125j), ('XZXIII', 0.0625), 
                    ('XZXIIZ', -0.03125), ('XZXZII', -0.03125), 
                    ('XZYIIZ', 0.03125j), ('XZYZII', -0.03125j), 
                    ('YXIIZI', -0.03125j), ('YXIZII', 0.03125j), 
                    ('YYIIII', (0.0625+0j)), ('YYIIZI', (-0.03125+0j)), 
                    ('YYIZII', (-0.03125+0j)), ('YZXIIZ', -0.03125j), 
                    ('YZXZII', 0.03125j), ('YZYIII', (0.0625+0j)), 
                    ('YZYIIZ', (-0.03125+0j)), ('YZYZII', (-0.03125+0j)), 
                    ('ZIIIII', 0.4375), ('ZIIZII', 0.0625)]
    h = pauli_sum(qubit_op)
    # a = pauli_sum(pauli_ops)
    
    
    n = 78
    t = 0
    # fn = flip_bit(n, t)
    # sn = swap_bits(4, 0, 1)
    # print(f'Flipping the {t}th bit of {n} gives {fn}')
    # print(f'Swapping the {0}th bit and the {1} bit of {4} gives {sn}')

    c = 2
    n = 8
    a = one_fixed_bit(n, c, False)
    for i in a:
        print("i:",i)
        if i[n-c-1] != '1':
            raise Exception('WRONG')
        
        i = int(i, 2)
        flipped_i = flip_bit(int(i), c) 
        print(f'{flipped_i:0{n}b}')
        if f'{flipped_i:0{n}b}'[-(c+1)] != '0':
            raise Exception('WRONG')





import pyscf
from pyscf import scf, gto
import numpy as np


def add_spin_one_body(operator):
    return np.kron(operator, np.eye(2))

def add_spin_two_body(operator, make_as=True):
    operator = np.kron(operator, np.einsum("pr, qs -> pqrs", np.eye(2), np.eye(2)))

    if make_as:
        operator = operator - operator.transpose(0,1,3,2)

    return operator


def get_hs(mol, is_rhf=False):
    L = mol.nao

    s = mol.intor_symmetric("int1e_ovlp")

    kinetic = mol.intor_symmetric("int1e_kin")
    potential = mol.intor_symmetric("int1e_nuc")
    h = kinetic + potential

    # "u" comes out as an (L**2, L**2) matrix. First we change it to (L,L,L,L) to fit the 4-index matrix element
    # additionally, it comes out in chemestry notation (ij|kl) = <ik|jl>, so swap axis 1 and 2 to get <ij|kl>
    u = mol.intor("int2e") 
    u = u.reshape(L, L, L, L)


    u = u.transpose(0, 2, 1, 3)
    if is_rhf:
        h, u = get_rhf(mol, h, u, s)


    # u = u.transpose(0, 2, 1, 3)

    # print(f"Before {h.shape = }")
    h = add_spin_one_body(h)
    # print(f"After  {h.shape = }")
    # # print(h)
    # print(f"Before {u.shape = }")
    u = add_spin_two_body(u)
    # print(f"After  {u.shape = }")

    return h, u


def get_rhf(mol,h,u,s):
    mf = scf.RHF(mol)
    mf.kernel()

    # print(mf.mo_coeff)
    c = mf.mo_coeff

    # print(s)

    somethingelse = np.einsum("pr,qs,pq -> rs", c, c, s)
    # print(somethingelse)

    U = np.einsum("pi,qj,rk,sl,pqrs -> ijkl", c, c, c, c, u, optimize=True)
    H = np.einsum("pr,qs,pq -> rs", c, c, h)

    U = U - U.transpose(0, 1, 3, 2)
    f = H - 0.5 * np.einsum("piqi -> pq",U)

    # print(f)

    return H, U



if __name__ == '__main__':

    # geometry = "He 0 0 0"
    #"Li 0 0 0; H 0 0 2.0"
    geometry = "H 0 0 0; H 0 0 0.7414"
    basis = "sto-3g"
    charge = 0

    mol = pyscf.gto.Mole()
    # mol.unit = "bohr" # Default is angstrom
    mol.build(atom=geometry, basis=basis, charge=charge)

    h, u = get_hs(mol, True)


    for i in range(4):
        for j in range(4):
            for k in range(4):
                for l in range(4):
                    if not np.allclose(u[i,j,k,l], 0):
                        print(f'{i}{j}{k}{l}, {u[i,j,k,l]}')
        
    # print(h)
    # print(u)
    # print(h.shape)
    # print(u.shape